searchNodes=[{"doc":"Carbonite implements the Change-Data-Capture pattern on top of PostgreSQL using database triggers. It keeps a central changes table where all mutations of participating tables are recorded. Each changes row is associated to a single row in the transactions table using PostgreSQL's internal transaction id as the foreign key. This leads to the following interesting properties: All changes created within a database transaction automatically and implicitly belong to the same record in the transactions table, even if they're created separatedly and agnostic of each other in the application logic. This gives the developer a &quot;natural&quot; way to group related changes into events (more on events later). As the changes table is associated to the transactions table via a non-nullable foreign key constraint, the entry in the transactions table must be created before any changes . Attempting to modify a versioned table without prior insertion into the transactions table will result in an error. The transactions table carries transactional metadata which can be set by the developer on creation. Consequently, much of Carbonite's logic lives in database functions and triggers. To get started, we need to create a migration using Ecto. ℹ️ Trigger vs. Write-Ahead-Log Existing solutions for CDC on top of a PostgreSQL database (e.g., Debezium ) often tail the &quot;Write-Ahead-Log&quot; instead of using database logic &amp; triggers to create change records. While this is likely more performant than using triggers, it makes it difficult to correlate changes on a transaction level as Carbonite does, and has different consistency guarantees. Carbonite makes this trade-off in favour of simplicity and universality. You should be able to run Carbonite's migrations on any hosted PostgreSQL instance without the need to tweak its configuration or install custom extensions before. Installing the schema &amp; triggers The following migration installs Carbonite into its &quot;default prefix&quot;, a PostgreSQL schema aptly called carbonite_default , and installs the change capture trigger for an exemplary table called rabbits (in the public schema). In a real-world scenario, you will most likely want to install the trigger for a set of tables and optionally split the audit trail into multiple partitions. See Carbonite.Migrations for more information on migrations. mix ecto.gen.migration InstallCarbonite # priv/repo/migrations/20210704201534_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do use Ecto.Migration def up do # Creates carbonite_default schema and tables. Carbonite.Migrations . install_schema ( ) # For each table that you want to capture changes of, you need to install the trigger. Carbonite.Migrations . install_trigger ( :rabbits ) end def down do # Remove all triggers before dropping the schema. Carbonite.Migrations . drop_trigger ( :rabbits ) # Drop the schema &amp; tables. Carbonite.Migrations . drop_schema ( ) end end Primary Key Columns To speed up version lookups for a specific record, Carbonite copies its primary key(s) to the table_pk column of the changes table. The table keeps an index on this column together with the table prefix and name. By default, Carbonite will try to copy the :id column of the source table. If your table does not have a primary key, has a primary key with a different name, or has a composite primary key, you can override this using the primary_key_columns option of Carbonite.Migrations.install_trigger/2 and Carbonite.Migrations.configure_trigger/2 . # Disable PK copying Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ ] ) # Different name Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ &quot;identifier&quot; ] ) # Composite PK Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ &quot;house&quot; , &quot;apartment_no&quot; ] ) Since the changes table keeps versions of a multitude of different source tables, primary keys are first cast to string (the table_pk column has type VARCHAR[] ). For composite primary keys, set the primary_key_columns option to an array as shown above. Each component of a compound primary key will be cast to string before the components are joined into the array. Excluded and Filtered Columns In case your table contains sensitive data or data otherwise undesirable for change capturing, you can exclude columns using the excluded_columns option. Excluded columns will not appear in the captured data. If an UPDATE on a table solely touches excluded columns, the entire UPDATE will not be recorded. Carbonite.Migrations . install_trigger ( :rabbits , excluded_columns : [ &quot;age&quot; ] ) If you still want to capture changes to a column (in the changed field), but don't need the exact data, you can make it a &quot;filtered&quot; column. These columns appear as &quot;[FILTERED]&quot; in the data field. Carbonite.Migrations . install_trigger ( :rabbits , filtered_columns : [ &quot;age&quot; ] ) If you forgot to exclude a column, you can reconfigure a trigger for a particular table using configure_trigger/2 : # in another migration Carbonite.Migrations . configure_trigger ( :rabbits , excluded_columns : [ &quot;age&quot; ] ) Partitioning the Audit Trail Carbonite can install its tables into multiple database schemas using the prefix option. You can use this feature to &quot;partition&quot; your captured data. Carbonite.Migrations . install_schema ( carbonite_prefix : &quot;carbonite_lagomorpha&quot; ) Carbonite.Migrations . install_trigger ( :rabbits , carbonite_prefix : &quot;carbonite_lagomorpha&quot; ) If desired, tables can participate in multiple partitions by adding multiple triggers on to them. Keep in mind that each partition will need to be processed and purged separately, resulting in multiple streams of change data in your external storage. Inserting a Transaction In your application logic, before modifying a versioned table like rabbits , you need to first create a Carbonite.Transaction record. With Ecto.Multi The easiest way to do so is using Carbonite.Multi.insert_transaction/2 within an Ecto.Multi operation: Ecto.Multi . new ( ) |&gt; Carbonite.Multi . insert_transaction ( %{ meta : %{ type : &quot;rabbit_inserted&quot; } } ) |&gt; Ecto.Multi . insert ( :rabbit , &amp; MyApp.Rabbit . create_changeset ( &amp;1 . params ) ) |&gt; MyApp.Repo . transaction ( ) As you can see, the Carbonite.Transaction is a great place to store metadata for the operation. A &quot;transaction type&quot; would be an obvious choice to categorize the transactions. A user_id would be a good candidate for an transaction log, as well. Building a changeset for manual insertion If you don't have the luxury of an Ecto.Multi , you can create a changeset for a Carbonite.Transaction using Carbonite.Transaction.changeset/1 : MyApp.Repo . transaction ( fn -&gt; %{ meta : %{ type : &quot;rabbit_inserted&quot; } } |&gt; Carbonite.Transaction . changeset ( ) |&gt; MyApp.Repo . insert! ( ) # ... end ) Setting metadata outside of the Transaction In case you do not have access to metadata you want to persist in the Carbonite.Transaction at the code site where you create it, you can use Carbonite.Transaction.put_meta/2 to store metadata in the process dictionary . This metadata is merged into the metadata given to Carbonite.Multi.insert_transaction/2 . # e.g., in a controller or plug Carbonite.Transaction . put_meta ( :user_id , ... ) Retrieving data Of course, persisting the audit trail is not an end in itself. At some point you will want to read the data back and make it accessible to the user. Carbonite.Query offers a small suite of helper functions that make it easier to query the database for Transaction and Change records. Fetching Changes The Carbonite.Query.changes/2 function constructs an Ecto.Query from a schema struct, loading all changes stored for the given source record. % MyApp.Rabbit { id : 1 } |&gt; Carbonite.Query . changes ( ) |&gt; MyApp.Repo . all ( ) Testing / Bypassing Carbonite One of Carbonite's key features is that it is virtually impossible to forget to record a change to a table (due to the trigger) or to forget to insert an enclosing Carbonite.Transaction beforehand (due to the foreign key constraint between changes and transactions ). However, in some circumstances it may be desirable to temporarily switch off change capturing. One such situation is the use of factories (e.g. ExMachina) inside your test suite: Inserting a transaction before each factory call quickly becomes cumbersome and will unnecessarily increase execution time. To bypass the capture trigger, Carbonite's trigger configuration provides a toggle mechanism consisting of two fields: mode and override_transaction_id . The former you set while installing the trigger on a table in a migration, while the latter allows to &quot;override&quot; whatever has been set at runtime, and only for the current transaction. If you are using Ecto's SQL sandbox for running transactional tests, this means the override is going to be active until the end of the test case. As a result, you have two options: Leave the mode at the default value of :capture and turn off capturing as needed by switching to &quot;override mode&quot;. This means for every test case where you do not care about change capturing, you explicitly disable the trigger before any database calls; for instance, in a ExUnit setup block. This approach has the benefit that you still capture all changes by default, and can't miss to test a code path that (in production) would require a Carbonite.Transaction . It is, however, still pretty expensive at ~1 additional SQL call per test case. Set the mode to :ignore on all triggers in your :test environment and instead selectively turn on capturing in test cases where you want to assert on the captured data. For instance, you can set the trigger mode in your migration based on the Mix environment. This approach is cheaper as it does not require any action in your tests by default. Yet you should make sure that you test all code paths that do mutate change-captured tables, in order to assert that each of these inserts a transaction as well. The following code snippet illustrates the second approach: # config/config.exs config :my_app , carbonite_mode : :capture # config/test.exs config :my_app , carbonite_mode : :ignore # priv/repo/migrations/000000000000_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do @mode Application . compile_env! ( :my_app , :carbonite_mode ) def up do # ... Carbonite.Migrations . install_trigger ( :rabbits , mode : @mode ) end end # test/support/carbonite_helpers.exs defmodule MyApp.CarboniteHelpers do def carbonite_override_mode ( _ ) do Ecto.Multi . new ( ) |&gt; Carbonite.Multi . override_mode ( ) |&gt; MyApp.Repo . transaction ( ) :ok end end # test/some_test.exs def SomeTest do use MyApp.DataCase import MyApp.CarboniteHelpers setup [ :carbonite_override_mode ] end","ref":"Carbonite.html","title":"Carbonite","type":"module"},{"doc":"Returns the default audit trail prefix.","ref":"Carbonite.html#default_prefix/0","title":"Carbonite.default_prefix/0","type":"function"},{"doc":"A Carbonite.Change records a mutation on a database table. INSERT statements lead to a Change where the data field contains the inserted row as a JSON object while the changed field is an empty list. UPDATE statements contain the updated record in data while the changed field is a list of attributes that have changed. DELETE statements have the delete data in data while changed is again an empty list.","ref":"Carbonite.Change.html","title":"Carbonite.Change","type":"module"},{"doc":"","ref":"Carbonite.Change.html#t:t/0","title":"Carbonite.Change.t/0","type":"type"},{"doc":"Functions to setup Carbonite audit trails in your migrations.","ref":"Carbonite.Migrations.html","title":"Carbonite.Migrations","type":"module"},{"doc":"Alters a triggers configuration for a given table. You need to specify all desired options each time you call this function, as the previous configuration will be replaced. Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot; primary_key_columns is a list of columns that form the primary key of the table ( defaults to ` [ &quot;id&quot; ] ` , set to ` [ ] ` to disable ) excluded_columns is a list of columns to exclude from change captures filtered_columns is a list of columns that appear as '[FILTERED]' in the data mode is either :capture or :ignore and defines the default behaviour of the trigger","ref":"Carbonite.Migrations.html#configure_trigger/2","title":"Carbonite.Migrations.configure_trigger/2","type":"function"},{"doc":"Removes a Carbonite audit trail from the database. Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#drop_schema/1","title":"Carbonite.Migrations.drop_schema/1","type":"function"},{"doc":"Removes a Carbonite audit trail from the database but keeps the schema. Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#drop_tables/1","title":"Carbonite.Migrations.drop_tables/1","type":"function"},{"doc":"Removes a change capture trigger from a table. Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#drop_trigger/2","title":"Carbonite.Migrations.drop_trigger/2","type":"function"},{"doc":"Installs a Carbonite audit trail. Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#install_schema/1","title":"Carbonite.Migrations.install_schema/1","type":"function"},{"doc":"Installs a change capture trigger on a table. Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot; primary_key_columns is a list of columns that form the primary key of the table ( defaults to ` [ &quot;id&quot; ] ` , set to ` [ ] ` to disable ) excluded_columns is a list of columns to exclude from change captures filtered_columns is a list of columns that appear as '[FILTERED]' in the data","ref":"Carbonite.Migrations.html#install_trigger/2","title":"Carbonite.Migrations.install_trigger/2","type":"function"},{"doc":"","ref":"Carbonite.Migrations.html#t:column_name/0","title":"Carbonite.Migrations.column_name/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:prefix/0","title":"Carbonite.Migrations.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:schema_option/0","title":"Carbonite.Migrations.schema_option/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:table_name/0","title":"Carbonite.Migrations.table_name/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:trigger_config_option/0","title":"Carbonite.Migrations.trigger_config_option/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:trigger_option/0","title":"Carbonite.Migrations.trigger_option/0","type":"type"},{"doc":"This module provides functions for dealing with audit trails in the context of Ecto.Multi.","ref":"Carbonite.Multi.html","title":"Carbonite.Multi","type":"module"},{"doc":"Adds an insert operation for a Carbonite.Transaction to an Ecto.Multi . Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; params map of params for the Carbonite.Transaction (e.g., :meta ) Multiple inserts in the same transaction Normally, you should have exactly one insert_transaction/3 call per database transaction. In practise, there are two scenarios in this function may be called multiple times: If an operation A, which calls insert_transaction/3 , sometimes is nested within an outer operation B, which also calls insert_transaction/3 . In tests using Ecto's SQL sandbox, subsequent calls to transactional operations (even to the same operation twice) are wrapped inside the overarching test transaction, and hence also effectively call insert_transaction/3 within the same transaction. While the first scenario can be resolved using appropriate control flow (e.g. by conditionally disabling the inner insert_transaction/3 call), the second scenario is quite common and often unavoidable. Therefore, insert_transaction/3 ignores subsequent calls within the same database transaction (equivalent to ON CONFLICT DO NOTHING ), discarding metadata passed to all calls but the first.","ref":"Carbonite.Multi.html#insert_transaction/3","title":"Carbonite.Multi.insert_transaction/3","type":"function"},{"doc":"Sets the current transaction to &quot;override mode&quot; for all tables in a translation log.","ref":"Carbonite.Multi.html#override_mode/2","title":"Carbonite.Multi.override_mode/2","type":"function"},{"doc":"","ref":"Carbonite.Multi.html#t:insert_transaction_option/0","title":"Carbonite.Multi.insert_transaction_option/0","type":"type"},{"doc":"","ref":"Carbonite.Multi.html#t:override_mode_option/0","title":"Carbonite.Multi.override_mode_option/0","type":"type"},{"doc":"","ref":"Carbonite.Multi.html#t:params/0","title":"Carbonite.Multi.params/0","type":"type"},{"doc":"","ref":"Carbonite.Multi.html#t:prefix/0","title":"Carbonite.Multi.prefix/0","type":"type"},{"doc":"Implements the outbox pattern to process (and evict) recorded transactions in the database, in order of insertion.","ref":"Carbonite.Outbox.html","title":"Carbonite.Outbox","type":"module"},{"doc":"Builds an Ecto.Multi that can be used to load Carbonite.Transaction records from database (in order of insertion) and process them one-by-one through a function. Processed items are marked as processed and can be purged. Options batch_size is the size of records to load in one chunk, defaults to 20 min_age is the minimum age of a record, defaults to 300 seconds (see below) prefix is the Carbonite schema, defaults to &quot;carbonite_default&quot; Long running transactions &amp; insertion order A warning: As Carbonite.Transaction records are inserted at the beginning of a database transaction, their inserted_at is already a bit in the past when they become visible to other connections, e.g. to your processing job. This means that in case of long running transactions, Carbonite.Transaction records with more recent inserted_at values might be processed before older ones, and hence the eventual total order of inserted_at in the processed records can not be guaranteed. To mitigate this issue, Carbonite will by default exclude records younger than min_age seconds from processing. Tweak this setting if you have even longer transactions.","ref":"Carbonite.Outbox.html#process/2","title":"Carbonite.Outbox.process/2","type":"function"},{"doc":"Builds an Ecto.Multi that can be used to delete Carbonite.Transaction records and their associated Carbonite.Change rows from the database once they have been successfully processed using process/2 . Options prefix is the Carbonite schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Outbox.html#purge/1","title":"Carbonite.Outbox.purge/1","type":"function"},{"doc":"","ref":"Carbonite.Outbox.html#t:prefix/0","title":"Carbonite.Outbox.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Outbox.html#t:process_fun/0","title":"Carbonite.Outbox.process_fun/0","type":"type"},{"doc":"","ref":"Carbonite.Outbox.html#t:process_option/0","title":"Carbonite.Outbox.process_option/0","type":"type"},{"doc":"","ref":"Carbonite.Outbox.html#t:purge_option/0","title":"Carbonite.Outbox.purge_option/0","type":"type"},{"doc":"This module provides query functions for retrieving audit trails from the database.","ref":"Carbonite.Query.html","title":"Carbonite.Query","type":"module"},{"doc":"Returns an Ecto.Query that can be used to select changes for a single record. Given an Ecto.Schema struct, this function builds a query that fetches all changes recorded for it from the database, ordered ascending by their ID (i.e., roughly by insertion date descending). Example % MyApp.Rabbit { id : 1 } |&gt; Carbonite.Query . changes ( ) |&gt; MyApp.Repo . all ( ) Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; table_prefix allows to override the table prefix, defaults to schema prefix of the record preload can be used to preload the transaction","ref":"Carbonite.Query.html#changes/2","title":"Carbonite.Query.changes/2","type":"function"},{"doc":"Returns an Ecto.Query that can be used to select or delete the &quot;current&quot; transaction. This function is useful when your tests run in a database transaction using Ecto's SQL sandbox. Example: Asserting on the current transaction When you insert your Carbonite.Transaction record somewhere inside your domain logic, you do not wish to return it to the caller only to be able to assert on its attributes in tests. This example shows how you could assert on the metadata inserted. # Test running inside Ecto&#39;s SQL sandbox. test &quot;my test&quot; do some_operation_with_a_transaction ( ) assert current_transaction_meta ( ) == %{ &quot;type&quot; =&gt; &quot;some_operation&quot; } end defp current_transaction_meta do Carbonite.Query . current_transaction ( ) |&gt; MyApp.Repo . one! ( ) |&gt; Map . fetch ( :meta ) end Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; preload can be used to preload the changes","ref":"Carbonite.Query.html#current_transaction/1","title":"Carbonite.Query.current_transaction/1","type":"function"},{"doc":"Returns an Ecto.Query that can be used to select transactions from the database. Examples Carbonite.Query . transactions ( ) |&gt; MyApp.Repo . all ( ) # Preload changes Carbonite.Query . transactions ( preload : :changes ) |&gt; MyApp.Repo . all ( ) # Same Carbonite.Query . transactions ( preload : true ) |&gt; MyApp.Repo . all ( ) Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; preload can be used to preload the changes","ref":"Carbonite.Query.html#transactions/1","title":"Carbonite.Query.transactions/1","type":"function"},{"doc":"","ref":"Carbonite.Query.html#t:changes_option/0","title":"Carbonite.Query.changes_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:current_transaction_option/0","title":"Carbonite.Query.current_transaction_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:prefix/0","title":"Carbonite.Query.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:preload/0","title":"Carbonite.Query.preload/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:transactions_option/0","title":"Carbonite.Query.transactions_option/0","type":"type"},{"doc":"A Carbonite.Transaction is the binding link between change records of tables. As such, it contains a set of optional metadata that describes the transaction.","ref":"Carbonite.Transaction.html","title":"Carbonite.Transaction","type":"module"},{"doc":"Builds a changeset for a new Carbonite.Transaction . The :meta map from the params will be merged with the metadata currently stored in the process dictionary.","ref":"Carbonite.Transaction.html#changeset/1","title":"Carbonite.Transaction.changeset/1","type":"function"},{"doc":"Returns the currently stored metadata.","ref":"Carbonite.Transaction.html#current_meta/0","title":"Carbonite.Transaction.current_meta/0","type":"function"},{"doc":"Stores a piece of metadata in the process dictionary. This can be useful in situations where you want to record a value at a system boundary (say, the user's account_id ) without having to pass it through to the database transaction. Returns the currently stored metadata.","ref":"Carbonite.Transaction.html#put_meta/2","title":"Carbonite.Transaction.put_meta/2","type":"function"},{"doc":"","ref":"Carbonite.Transaction.html#t:meta/0","title":"Carbonite.Transaction.meta/0","type":"type"},{"doc":"","ref":"Carbonite.Transaction.html#t:t/0","title":"Carbonite.Transaction.t/0","type":"type"},{"doc":"A Carbonite.Trigger stores per table configuration for the change capture trigger.","ref":"Carbonite.Trigger.html","title":"Carbonite.Trigger","type":"module"},{"doc":"","ref":"Carbonite.Trigger.html#t:mode/0","title":"Carbonite.Trigger.mode/0","type":"type"},{"doc":"","ref":"Carbonite.Trigger.html#t:t/0","title":"Carbonite.Trigger.t/0","type":"type"},{"doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS APPENDIX: How to apply the Apache License to your work. To apply the Apache License to your work, attach the following boilerplate notice, with the fields enclosed by brackets &quot;[]&quot; replaced with your own identifying information. (Don't include the brackets!) The text should be enclosed in the appropriate comment syntax for the file format. We also recommend that a file or class name and description of purpose be included on the same &quot;printed page&quot; as the copyright notice for easier identification within third-party archives. Copyright 2019 Parker Selbert Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"license.html","title":"LICENSE","type":"extras"},{"doc":"Carbonite makes it easy to create audit trails for tables in a PostgreSQL database and integrate them into an Elixir application. Carbonite implements the Change-Data-Capture pattern on top of a PostgreSQL database. It uses triggers to automatically record all changes applied to a database table in order to guarantee a complete audit trail of the contained data. It is centered around the idea that the database transaction is the natural auditing unit of data mutation. Any mutation on a participating table requires the developer first to record the operation metadata within the same database transaction. The metadata record is associated to the table change records by a constraint. On top of its database layer, Carbonite provides an API allowing developers to effortlessly retrieve, process, and purge the audit trails within the Elixir application. 🚧 This is work-in-progress and not yet ready to be used 🚧","ref":"readme.html","title":"README","type":"extras"},{"doc":"Convenient installation using migration functions Guaranteed consistency based on foreign-key constraints and ACID No mutation without recorded Change No Change without Transaction Customizable audit metadata (per transaction) Clear query-based interfaces to the audit trail Optional processing &amp; purging logic following the Outbox pattern Based on Ecto and Postgrex with no further dependencies. No configuration of PostgreSQL database needed","ref":"readme.html#features","title":"README - Features","type":"extras"},{"doc":"Requirements Due to its use of pg_current_xact_id , Carbonite requires ⚠️ PostgreSQL version 13 or above ⚠️. If you see an error message like the following, your PostgreSQL installation is too old: ** (Postgrex.Error) ERROR 42704 (undefined_object) type &quot;xid8&quot; does not exist Hex dependency # mix.exs def deps do [ { :carbonite , &quot;~&gt; 0.3.1&quot; } ] end","ref":"readme.html#installation","title":"README - Installation","type":"extras"},{"doc":"Carbonite implements the Change-Data-Capture pattern on top of PostgreSQL using database triggers. It keeps a central changes table where all mutations of participating tables are recorded. Each changes row is associated to a single row in the transactions table using PostgreSQL's internal transaction id as the foreign key. This leads to the following interesting properties: All changes created within a database transaction automatically and implicitly belong to the same record in the transactions table, even if they're created separatedly and agnostic of each other in the application logic. This gives the developer a &quot;natural&quot; way to group related changes into events (more on events later). As the changes table is associated to the transactions table via a non-nullable foreign key constraint, the entry in the transactions table must be created before any changes . Attempting to modify a versioned table without prior insertion into the transactions table will result in an error. The transactions table carries transactional metadata which can be set by the developer on creation. Consequently, much of Carbonite's logic lives in database functions and triggers. To get started, we need to create a migration using Ecto. ℹ️ Trigger vs. Write-Ahead-Log Existing solutions for CDC on top of a PostgreSQL database (e.g., Debezium ) often tail the &quot;Write-Ahead-Log&quot; instead of using database logic &amp; triggers to create change records. While this is likely more performant than using triggers, it makes it difficult to correlate changes on a transaction level as Carbonite does, and has different consistency guarantees. Carbonite makes this trade-off in favour of simplicity and universality. You should be able to run Carbonite's migrations on any hosted PostgreSQL instance without the need to tweak its configuration or install custom extensions before.","ref":"readme.html#usage","title":"README - Usage","type":"extras"},{"doc":"The following migration installs Carbonite into its &quot;default prefix&quot;, a PostgreSQL schema aptly called carbonite_default , and installs the change capture trigger for an exemplary table called rabbits (in the public schema). In a real-world scenario, you will most likely want to install the trigger for a set of tables and optionally split the audit trail into multiple partitions. See Carbonite.Migrations for more information on migrations. mix ecto.gen.migration InstallCarbonite # priv/repo/migrations/20210704201534_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do use Ecto.Migration def up do # Creates carbonite_default schema and tables. Carbonite.Migrations . install_schema ( ) # For each table that you want to capture changes of, you need to install the trigger. Carbonite.Migrations . install_trigger ( :rabbits ) end def down do # Remove all triggers before dropping the schema. Carbonite.Migrations . drop_trigger ( :rabbits ) # Drop the schema &amp; tables. Carbonite.Migrations . drop_schema ( ) end end Primary Key Columns To speed up version lookups for a specific record, Carbonite copies its primary key(s) to the table_pk column of the changes table. The table keeps an index on this column together with the table prefix and name. By default, Carbonite will try to copy the :id column of the source table. If your table does not have a primary key, has a primary key with a different name, or has a composite primary key, you can override this using the primary_key_columns option of Carbonite.Migrations.install_trigger/2 and Carbonite.Migrations.configure_trigger/2 . # Disable PK copying Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ ] ) # Different name Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ &quot;identifier&quot; ] ) # Composite PK Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ &quot;house&quot; , &quot;apartment_no&quot; ] ) Since the changes table keeps versions of a multitude of different source tables, primary keys are first cast to string (the table_pk column has type VARCHAR[] ). For composite primary keys, set the primary_key_columns option to an array as shown above. Each component of a compound primary key will be cast to string before the components are joined into the array. Excluded and Filtered Columns In case your table contains sensitive data or data otherwise undesirable for change capturing, you can exclude columns using the excluded_columns option. Excluded columns will not appear in the captured data. If an UPDATE on a table solely touches excluded columns, the entire UPDATE will not be recorded. Carbonite.Migrations . install_trigger ( :rabbits , excluded_columns : [ &quot;age&quot; ] ) If you still want to capture changes to a column (in the changed field), but don't need the exact data, you can make it a &quot;filtered&quot; column. These columns appear as &quot;[FILTERED]&quot; in the data field. Carbonite.Migrations . install_trigger ( :rabbits , filtered_columns : [ &quot;age&quot; ] ) If you forgot to exclude a column, you can reconfigure a trigger for a particular table using configure_trigger/2 : # in another migration Carbonite.Migrations . configure_trigger ( :rabbits , excluded_columns : [ &quot;age&quot; ] ) Partitioning the Audit Trail Carbonite can install its tables into multiple database schemas using the prefix option. You can use this feature to &quot;partition&quot; your captured data. Carbonite.Migrations . install_schema ( carbonite_prefix : &quot;carbonite_lagomorpha&quot; ) Carbonite.Migrations . install_trigger ( :rabbits , carbonite_prefix : &quot;carbonite_lagomorpha&quot; ) If desired, tables can participate in multiple partitions by adding multiple triggers on to them. Keep in mind that each partition will need to be processed and purged separately, resulting in multiple streams of change data in your external storage.","ref":"readme.html#installing-the-schema-triggers","title":"README - Installing the schema &amp; triggers","type":"extras"},{"doc":"In your application logic, before modifying a versioned table like rabbits , you need to first create a Carbonite.Transaction record. With Ecto.Multi The easiest way to do so is using Carbonite.Multi.insert_transaction/2 within an Ecto.Multi operation: Ecto.Multi . new ( ) |&gt; Carbonite.Multi . insert_transaction ( %{ meta : %{ type : &quot;rabbit_inserted&quot; } } ) |&gt; Ecto.Multi . insert ( :rabbit , &amp; MyApp.Rabbit . create_changeset ( &amp;1 . params ) ) |&gt; MyApp.Repo . transaction ( ) As you can see, the Carbonite.Transaction is a great place to store metadata for the operation. A &quot;transaction type&quot; would be an obvious choice to categorize the transactions. A user_id would be a good candidate for an transaction log, as well. Building a changeset for manual insertion If you don't have the luxury of an Ecto.Multi , you can create a changeset for a Carbonite.Transaction using Carbonite.Transaction.changeset/1 : MyApp.Repo . transaction ( fn -&gt; %{ meta : %{ type : &quot;rabbit_inserted&quot; } } |&gt; Carbonite.Transaction . changeset ( ) |&gt; MyApp.Repo . insert! ( ) # ... end ) Setting metadata outside of the Transaction In case you do not have access to metadata you want to persist in the Carbonite.Transaction at the code site where you create it, you can use Carbonite.Transaction.put_meta/2 to store metadata in the process dictionary . This metadata is merged into the metadata given to Carbonite.Multi.insert_transaction/2 . # e.g., in a controller or plug Carbonite.Transaction . put_meta ( :user_id , ... )","ref":"readme.html#inserting-a-transaction","title":"README - Inserting a Transaction","type":"extras"},{"doc":"Of course, persisting the audit trail is not an end in itself. At some point you will want to read the data back and make it accessible to the user. Carbonite.Query offers a small suite of helper functions that make it easier to query the database for Transaction and Change records. Fetching Changes The Carbonite.Query.changes/2 function constructs an Ecto.Query from a schema struct, loading all changes stored for the given source record. % MyApp.Rabbit { id : 1 } |&gt; Carbonite.Query . changes ( ) |&gt; MyApp.Repo . all ( )","ref":"readme.html#retrieving-data","title":"README - Retrieving data","type":"extras"},{"doc":"One of Carbonite's key features is that it is virtually impossible to forget to record a change to a table (due to the trigger) or to forget to insert an enclosing Carbonite.Transaction beforehand (due to the foreign key constraint between changes and transactions ). However, in some circumstances it may be desirable to temporarily switch off change capturing. One such situation is the use of factories (e.g. ExMachina) inside your test suite: Inserting a transaction before each factory call quickly becomes cumbersome and will unnecessarily increase execution time. To bypass the capture trigger, Carbonite's trigger configuration provides a toggle mechanism consisting of two fields: mode and override_transaction_id . The former you set while installing the trigger on a table in a migration, while the latter allows to &quot;override&quot; whatever has been set at runtime, and only for the current transaction. If you are using Ecto's SQL sandbox for running transactional tests, this means the override is going to be active until the end of the test case. As a result, you have two options: Leave the mode at the default value of :capture and turn off capturing as needed by switching to &quot;override mode&quot;. This means for every test case where you do not care about change capturing, you explicitly disable the trigger before any database calls; for instance, in a ExUnit setup block. This approach has the benefit that you still capture all changes by default, and can't miss to test a code path that (in production) would require a Carbonite.Transaction . It is, however, still pretty expensive at ~1 additional SQL call per test case. Set the mode to :ignore on all triggers in your :test environment and instead selectively turn on capturing in test cases where you want to assert on the captured data. For instance, you can set the trigger mode in your migration based on the Mix environment. This approach is cheaper as it does not require any action in your tests by default. Yet you should make sure that you test all code paths that do mutate change-captured tables, in order to assert that each of these inserts a transaction as well. The following code snippet illustrates the second approach: # config/config.exs config :my_app , carbonite_mode : :capture # config/test.exs config :my_app , carbonite_mode : :ignore # priv/repo/migrations/000000000000_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do @mode Application . compile_env! ( :my_app , :carbonite_mode ) def up do # ... Carbonite.Migrations . install_trigger ( :rabbits , mode : @mode ) end end # test/support/carbonite_helpers.exs defmodule MyApp.CarboniteHelpers do def carbonite_override_mode ( _ ) do Ecto.Multi . new ( ) |&gt; Carbonite.Multi . override_mode ( ) |&gt; MyApp.Repo . transaction ( ) :ok end end # test/some_test.exs def SomeTest do use MyApp.DataCase import MyApp.CarboniteHelpers setup [ :carbonite_override_mode ] end","ref":"readme.html#testing-bypassing-carbonite","title":"README - Testing / Bypassing Carbonite","type":"extras"},{"doc":"Inspiration The trigger-based table versioning derives from audit_trigger_91plus , an &quot;example of a generic trigger function&quot; hosted in the PostgreSQL wiki. Artwork The amazing Carbonite logo has been designed by Petra Herberger . Thank you! 💜","ref":"readme.html#acknowledgements","title":"README - Acknowledgements","type":"extras"},{"doc":"Copyright © 2021, Bitcrowd GmbH. You may use and redistribute Carbonite and its source code under the Apache 2.0 License, see the LICENSE file for details.","ref":"readme.html#copyright-and-license","title":"README - Copyright and License","type":"extras"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Added table_prefix option to Query.changes/2 allows to override schema prefix of given record Query.transactions/1 query selects all transactions","ref":"changelog.html#0-3-1-2021-10-23","title":"Changelog - [0.3.1] - 2021-10-23","type":"extras"},{"doc":"Added Carbonite.Migrations.drop_tables/1 allows to drop the carbonite audit trail without removing the schema Changed Renamed the option that can be passed to Carbonite.Migrations.drop_schema/1 from prefix to carbonite_prefix Changed Carbonite.Migrations.drop_schema/1 to also drop the tables Made Carbonite.Multi.insert_transaction/3 ignore conflicting INSERT s within the same transaction Also, changed Carbonite.Multi.insert_transaction/3 to always reloads all fields from the database after insertion, immediately returning the JSONinified meta payload Fixed Fixed ignore mode when override_transaction_id is NULL","ref":"changelog.html#0-3-0-2021-10-22","title":"Changelog - [0.3.0] - 2021-10-22","type":"extras"},{"doc":"Fixed Fixed broken documentation","ref":"changelog.html#0-2-1-2021-10-10","title":"Changelog - [0.2.1] - 2021-10-10","type":"extras"},{"doc":"Added Store primary key(s) on changes table and index them Add Carbonite.Query module current_transaction/2 allows to fetch the ongoing transaction (for sandbox tests) changes/2 allows to fetch the changes of an invidual source record Update Postgrex to 0.15.11 and drop local Xid8 type Add mode field to trigger (capture or ignore) Add &quot;override mode&quot; reversing the mode option for the current transaction to enable/disable capturing on demand (e.g. in tests) Add filtered columns Changed Moved top-level functions to nested modules Transaction and Multi Made table_pk be NULL when primary_key_columns is an empty array Default primary_key_columns to [&quot;id&quot;] Renamed prefix option to carbonite_prefix on install_schema/2 for consistency","ref":"changelog.html#0-2-0-2021-10-10","title":"Changelog - [0.2.0] - 2021-10-10","type":"extras"},{"doc":"Initial release.","ref":"changelog.html#0-1-0-2021-09-01","title":"Changelog - [0.1.0] - 2021-09-01","type":"extras"}]